# This module remains as a legacy module for the time being. It will be removed in
# the future.

# import re
# from typing import Dict, List, Optional, Union
#
# from pydantic import validator
#
# from osw.model.static import OswBaseModel
# from osw.utils.strings import *
